const std = @import("std");
const sdl = @import("sdl");
const cfg = @import("cfg.zig");
const R = @import("R.zig");
const PixelGrid = @import("grid.zig").Grid(Pixel);

const Self = @This();

renderer: sdl.Renderer,
grid: PixelGrid,
texture: sdl.Texture,

pub const Pixel = u1;

pub const Command = enum {
    off,
    on,
    invert,

    fn paint(self: Command, p: Pixel) Pixel {
        return switch (self) {
            .off => 0,
            .on => 1,
            .invert => if (p == 0) 1 else 0,
        };
    }

    fn transform(self: Command, p: *Pixel) void {
        p.* = self.paint(p.*);
    }
};

pub const Axis = enum { x, y, d };

pub fn init(
    renderer: sdl.Renderer,
    allocator: std.mem.Allocator,
    width: ?u31,
    height: ?u31,
) !Self {
    const grid = try PixelGrid.init(
        allocator,
        width orelse cfg.dwidth,
        height orelse cfg.dheight,
    );
    errdefer grid.deinit();
    grid.fill(0);

    const texture = try sdl.createTexture(
        renderer,
        .rgb888,
        .streaming,
        width orelse cfg.dwidth,
        height orelse cfg.dheight,
    );
    errdefer texture.destroy();

    return .{
        .renderer = renderer,
        .grid = grid,
        .texture = texture,
    };
}

pub fn deinit(self: Self) void {
    self.texture.destroy();
    self.grid.deinit();
}

pub fn present(self: Self) !void {
    {
        var pixels = try self.texture.lock(null);
        defer pixels.release();

        for (0..self.grid.height()) |y| {
            const line = pixels.scanline(y, u8);
            for (0..self.grid.width()) |x| {
                const v = self.grid.at(x, y).?.*;
                @memset(line[x * 4 .. x * 4 + 3], @as(u8, v) * 255);
            }
        }
    }

    try self.renderer.copy(self.texture, null, null);
}

pub fn point(self: Self, c: Command, x: usize, y: usize) void {
    c.transform(self.grid.at(x, y) orelse return);
}

pub fn fill(self: Self, c: Command) void {
    for (0..self.grid.height()) |y| {
        for (0..self.grid.width()) |x| {
            self.point(c, x, y);
        }
    }
}

pub fn copy(self: Self, pb: Self, c: Command, ox: i32, oy: i32) void {
    for (0..pb.grid.height()) |y| {
        if (-oy > y) continue;
        const dy: usize = @intCast(@as(i32, @intCast(y)) + oy);
        if (dy >= self.grid.height()) continue;
        for (0..pb.grid.width()) |x| {
            if (-ox > x) continue;
            const dx: usize = @intCast(@as(i32, @intCast(x)) + ox);
            if (dx >= self.grid.width()) continue;
            if (pb.grid.at(x, y).?.* != 0)
                self.point(c, dx, dy);
        }
    }
}

pub fn flip(self: Self, axis: Axis) void {
    std.debug.assert(self.grid.isSquare());
    const size = self.grid.width();
    switch (axis) {
        .x => {
            for (0..size) |y| {
                for (0..size / 2) |x| {
                    const tmp = self.grid.at(x, y).?.*;
                    self.grid.at(x, y).?.* = self.grid.at(size - x - 1, y).?.*;
                    self.grid.at(size - x - 1, y).?.* = tmp;
                }
            }
        },
        .y => {
            for (0..size / 2) |y| {
                for (0..size) |x| {
                    const tmp = self.grid.at(x, y).?.*;
                    self.grid.at(x, y).?.* = self.grid.at(x, size - y - 1).?.*;
                    self.grid.at(x, size - y - 1).?.* = tmp;
                }
            }
        },
        .d => {
            for (0..size) |y| {
                for (y..size) |x| {
                    const tmp = self.grid.at(x, y).?.*;
                    self.grid.at(x, y).?.* = self.grid.at(y, x).?.*;
                    self.grid.at(y, x).?.* = tmp;
                }
            }
        },
    }
}

pub fn rotate(self: Self, quarters: i32) void {
    switch (@mod(quarters, 4)) {
        else => {},
        1 => { // right
            self.flip(.d);
            self.flip(.x);
        },
        2 => { // tupsy-turvy
            self.rotate(1);
            self.rotate(1);
        },
        3 => { // left
            self.flip(.d);
            self.flip(.y);
        },
    }
}
