const std = @import("std");
const sdl = @import("sdl");
const builtin = @import("builtin");
const cfg = @import("cfg.zig");
const Resource = @import("Resource.zig");
const num_scancodes = sdl.c.SDL_NUM_SCANCODES;
const num_buttons = sdl.c.SDL_CONTROLLER_BUTTON_MAX;

const Self = @This();
pub const black = sdl.Color{ .r = 0, .g = 0, .b = 0, .a = 255 };
pub const white = sdl.Color{ .r = 255, .g = 255, .b = 255, .a = 255 };

allocator: std.mem.Allocator,
window: sdl.Window,
renderer: sdl.Renderer,
shouldQuit: bool = false,
scstates: [num_scancodes]InputState = .{.up} ** num_scancodes,
mbstates: [256]InputState = .{.up} ** 256,
gcstates: [num_buttons]InputState = .{.up} ** num_buttons,
mouse: Point = .{ .x = 0, .y = 0 },
resources: std.ArrayList(Resource),
fps: u64,
nextTime: u64,
tick: u64 = 0,
fullscreen: bool = false,

pub const ResourceID = ?u31;

pub const InitArgs = struct {
    title: [:0]const u8,
    width: usize,
    height: usize,
    scale: usize,
    fps: u32,
    chdir: bool,
    windowDim: sdl.WindowFlags.Dimension = .default,
};

pub const InputState = enum {
    up,
    down,
    release,
    press,

    pub fn next(self: InputState) InputState {
        return switch (self) {
            .release => .up,
            .press => .down,
            else => self,
        };
    }

    pub fn isDown(self: InputState) bool {
        return switch (self) {
            .down, .press => true,
            else => false,
        };
    }

    pub fn isUp(self: InputState) bool {
        return switch (self) {
            .up, .release => true,
            else => false,
        };
    }
};

pub const Point = struct {
    x: i32,
    y: i32,
};

pub fn init(allocator: std.mem.Allocator, comptime args: InitArgs) !Self {
    try sdl.init(.{ .video = true, .events = true });
    errdefer sdl.quit();

    // chdir to basepath
    if (args.chdir and builtin.os.tag != .windows) {
        const argv = try std.process.argsAlloc(allocator);
        defer std.process.argsFree(allocator, argv);
        if (argv.len <= 1) {
            const basepath = sdl.c.SDL_GetBasePath();
            defer if (basepath != null) sdl.c.SDL_free(basepath);
            if (basepath != null) {
                std.log.debug("basepath: {s}", .{basepath});
                try std.os.chdir(std.mem.span(basepath));
            } else {
                std.log.warn("{s}", .{sdl.c.SDL_GetError()});
            }
        } else {
            try std.os.chdir(argv[1]);
        }
    }

    var window = try sdl.createWindow(
        args.title,
        .default,
        .default,
        args.width * args.scale,
        args.height * args.scale,
        .{ .dim = args.windowDim, .resizable = true },
    );
    errdefer window.destroy();

    var renderer = try sdl.createRenderer(
        window,
        null,
        .{ .accelerated = true },
    );
    errdefer renderer.destroy();
    try renderer.setLogicalSize(cfg.dwidth, cfg.dheight);

    var resources = std.ArrayList(Resource).init(allocator);
    errdefer resources.deinit();

    return .{
        .allocator = allocator,
        .window = window,
        .renderer = renderer,
        .resources = resources,
        .fps = args.fps,
        .nextTime = if (args.fps > 0) sdl.getTicks64() else undefined,
    };
}

pub fn deinit(self: Self) void {
    for (self.resources.items) |e| e.deinit();
    self.resources.deinit();
    self.renderer.destroy();
    self.window.destroy();
    sdl.quit();
}

pub fn cycleEvents(self: *Self) void {
    for (&self.scstates) |*v| v.* = v.next();
    for (&self.mbstates) |*v| v.* = v.next();
    for (&self.gcstates) |*v| v.* = v.next();

    while (sdl.pollEvent()) |e| {
        switch (e) {
            .quit => self.shouldQuit = true,
            .key_down => |k| {
                if (!k.is_repeat)
                    self.scstates[@intFromEnum(k.scancode)] = .press;
            },
            .key_up => |k| {
                if (!k.is_repeat)
                    self.scstates[@intFromEnum(k.scancode)] = .release;
            },
            .mouse_button_down => |m| {
                self.mbstates[@intFromEnum(m.button)] = .press;
            },
            .mouse_button_up => |m| {
                self.mbstates[@intFromEnum(m.button)] = .release;
            },
            .mouse_motion => |m| {
                self.mouse.x = m.x;
                self.mouse.y = m.y;
            },
            else => {},
        }
    }
}

pub fn drawBegin(self: Self) !void {
    _ = self;
}

pub fn drawEnd(self: *Self) !void {
    if (self.fps > 0) {
        self.nextTime += 1000 / self.fps;
        if (self.fps > 0 and self.tick % 3 < 2)
            self.nextTime += 1;
        const currentTime = sdl.getTicks64();
        if (self.nextTime > currentTime)
            sdl.delay(@truncate(self.nextTime - currentTime));
    }

    self.renderer.present();
    self.tick += 1;
}

pub fn toggleFullscreen(self: *Self) void {
    self.fullscreen = !self.fullscreen;
    _ = sdl.c.SDL_SetWindowFullscreen(
        self.window.ptr,
        if (self.fullscreen)
            sdl.c.SDL_WINDOW_FULLSCREEN_DESKTOP
        else
            0,
    );
}

pub fn resource(self: *Self, tag: Resource.Tag, path: [:0]const u8) !ResourceID {
    const hash = std.hash.Wyhash.hash(0, std.mem.sliceTo(path, 0));
    for (self.resources.items, 0..) |e, i|
        if (e.hash == hash)
            return @truncate(i);

    var resour = Resource.load(self.*, tag, path) catch return null;
    errdefer resour.deinit();
    resour.hash = hash;

    try self.resources.append(resour);
    return @truncate(self.resources.items.len - 1);
}

pub const res = resource;

pub fn img(self: *Self, path: [:0]const u8) !ResourceID {
    return self.resource(.image, path);
}
